from simpy import Environment
from typing import Callable
from simso.core.Processor import ProcInfo
from simso.core.Monitor import Monitor
from simso.core.FaultEvent import ProcessorFailureEvent, VoltageChangeEvent, TimingChangeEvent
from simso.core.JobEvent import JobEvent


class FaultFunction:
    """
    The FaultFunction class changes the actual behaviour of the faulty element.
    """

    def __init__(self, start_function: Callable, end_function: Callable):
        self.start_function = start_function
        self.end_function = end_function


class FaultInfo:
    """
    The FaultType class is responsible for causing the error.
    """

    def __init__(self, fault_type: str, start_time: float, end_time: float,
                 fault_target, value, parameter: str):
        self.fault_type = fault_type
        self.start_time = start_time
        self.end_time = end_time
        self.fault_target = fault_target
        self.value = value
        self.parameter = parameter


class Fault:
    """
    The Fault base class is the simpy process for triggering a certain error.
    """

    def __init__(self, sim, fault_info: FaultInfo,  start_function: Callable, end_function: Callable):
        self._start_function = start_function
        self._end_function = end_function
        self._fault_info = fault_info
        self._sim = sim
        self._monitor = Monitor(name="Monitor" + fault_info.fault_type,
                                env=sim)
        self.process = self._sim.process(self.release())

    @property
    def fault_info(self):
        return self._fault_info

    @property
    def duration_ms(self):
        return self._fault_info.end_time - self._fault_info.start_time

    @property
    def duration_cycles(self):
        return self.duration_ms * self._sim.cycles_per_ms

    def release(self):
        yield self._sim.timeout(self._fault_info.start_time * self._sim.cycles_per_ms)

        self._start_function(self._sim)

        yield self._sim.timeout((self._fault_info.end_time -
                                 self._fault_info.start_time) * self._sim.cycles_per_ms)

        self._end_function(self._sim)


# TODO: Does not work when only 1 processor in model, maybe use speed = 0 ?
class ProcessorFailure(Fault):
    """
    The ProcessorFailure class simulates a complete stoppage of a processor.
    """
    fields = ['fault_target']

    def __init__(self, sim, fault_info):
        super().__init__(sim, fault_info, self.stop_processor, self.start_processor)
        self._processor = None

    def stop_processor(self, model):
        self._processor = next(
            p for p in model.processors if p.identifier == self._fault_info.fault_target.identifier)
        self._processor.fail(self)
        self._monitor.observe(ProcessorFailureEvent(self))

    def start_processor(self, model):
        self._processor.fail()


class VoltageChange(Fault):

    fields = ['fault_target', 'value']

    def __init__(self, sim, fault_info):
        super().__init__(sim, fault_info, self.start_voltage_change, self.end_voltage_change)
        self._processor = None

    def start_voltage_change(self, model):
        self._processor = next(
            p for p in model.processors if p.identifier == self._fault_info.fault_target.identifier)
        self._processor.set_speed(
            self._fault_info.value * self._processor.speed)
        self._monitor.observe(VoltageChangeEvent(self))

    def end_voltage_change(self, model):
        self._processor.set_speed(self._fault_info.fault_target.speed)


class VoltageDrop(VoltageChange):
    """
    The VoltageDrop class simulates the slowdown of one or more processors due to voltage drops.
    """

    def __init__(self, sim, fault_info):
        assert fault_info.value >= 0 and fault_info.value <= 1.0, "Voltage drop reduces the speed of a processor, i.e. scaling has to be >= 0 and <= 1"
        super().__init__(sim, fault_info)


# TODO: rethink this class
class PriorityInversion(FaultFunction):
    """
    The PriorityInversion class simulates changes in the priority of a job.
    """

    def __init__(self, proc_info, until_sched: bool = False):
        """
        lambd: Expects task and new value as parameter and returns old value
        """
        super().__init__(self.start_priority_inversion, self.end_priority_inversion)
        self._proc_info = proc_info
        self._until_sched = until_sched

    def start_priority_inversion(self, model):
        proc = next(p for p in model.processors if p.identifier ==
                    self._proc_info.identifier)
        sched = model.scheduler
        running_jobs = [
            p.running for p in model.processors if p.running is not None]
        job = next(t.job for t in sched.task_list if t.job.is_active()
                   and t.job not in running_jobs)

        if job is not None:
            pass

    def end_priority_inversion(self, model):
        pass


class TimingChange(Fault):
    """
    The TimingError class simulates changes in the timing parameters, e.g. deadline, period etc.
    """

    fields = ['fault_target', 'parameter', 'value']

    def __init__(self, sim, fault_info):
        super().__init__(sim, fault_info, self.start_timing_change, self.end_timing_change)
        self._old_value = getattr(
            self._fault_info.fault_target, fault_info.parameter)

    def start_timing_change(self, model):
        setattr(self._fault_info.fault_target, self._fault_info.parameter,
                self._fault_info.value)
        self._monitor.observe(TimingChangeEvent(self))

    def end_timing_change(self, model):
        setattr(self._fault_info.fault_target,
                self._fault_info.parameter, self._old_value)


class TimingError(TimingChange):
    """
    The TimingError class simulates changes in the timing parameters, e.g. deadline, period etc.
    """

    def __init__(self, sim, fault_info):
        super().__init__(sim, fault_info)


class Redundancy(TimingChange):

    def __init__(self, sim, fault_info):
        fault_info.parameter = "wcet"
        fault_info.value = fault_info.fault_target.wcet + fault_info.value
        super().__init__(sim, fault_info)


fault_types = {
    "TimingError": TimingError,
    "TimingChange": TimingChange,
    "Redundancy": Redundancy,
    "VoltageChange": VoltageChange,
    "VoltageDrop": VoltageDrop,
    "ProcessorFailure": ProcessorFailure
}

fault_types_names = ["TimingError",  "VoltageDrop",  "ProcessorFailure"]
# "TimingChange", "Redundancy", "VoltageChange",

def FaultFactory(sim, fault_info):
    """
    Task factory. Return and instantiate the correct class according to the
    task_info.
    """
    return fault_types[fault_info.fault_type](sim, fault_info)
